Attribute VB_Name = "StringUtils"
'-----------------------------------------------------------------------------
' Utilities from http://blog.nkadesign.com/microsoft-access/
' (c) Renaud Bompuis, 2013
' Licensed under the Creative Commons Attribution License
' http://creativecommons.org/licenses/by/3.0/
' http://creativecommons.org/licenses/by/3.0/legalcode
'
' Free for re-use in any application or tutorial providing clear credit
' is made about the origin of the code and a link to the site above
' is prominently displayed where end-user can access it.
'-----------------------------------------------------------------------------
Option Compare Database
Option Explicit

'-----------------------------------------------------------------------------
' Replaces the format item in a specified string with the string representation
' of a corresponding object in a specified array.
' Identical to String.Format(string format, params object[] args) in .Net
' See http://msdn.microsoft.com/en-us/library/system.string.format.aspx
'-----------------------------------------------------------------------------
Public Function sFormat(ByRef frmt As String, ParamArray args() As Variant) As String
    Static m_StrBuilder As Object
    If m_StrBuilder Is Nothing Then Set m_StrBuilder = CreateObject("System.Text.StringBuilder")
    
    If UBound(args) = -1 Then
        ' Wrong number of arguments
        Err.Raise 450, "sFormat()"
        Exit Function
    End If
    
    ' Unpack args if they were passed verbatim from another ParamArray
    Dim params() As Variant
    If (UBound(args) - LBound(args) = 0) And (VarType(args(0)) = (vbArray + vbVariant)) Then
        params = args(0)
    Else
        params = args
    End If
    
    ' Double check that all the parameters are of a type that we can handle otherwise convert to base type
    ' Unescape parameter strings, just in case they contain some haracter litterals or unicode sequences
    Dim i As Long
    Dim p As Variant
    For i = LBound(params) To UBound(params)
        p = ConvertVariantToBaseType(params(i))
        If VarType(p) = VbVarType.vbString Then params(i) = UnEscStr(p) Else params(i) = p
    Next i
    
    ' Clear and roughly adjust the capacity of the string builder
    m_StrBuilder.Length = 0
    m_StrBuilder.Capacity = Len(frmt) + (UBound(params) + 1) * 8
    ' Construct the final string
    sFormat = m_StrBuilder.AppendFormat_4(UnEscStr(frmt), params).ToString()
    m_StrBuilder.Length = 0
End Function


'-----------------------------------------------------------------------------
' Takes a string that may contain .Net escaped characters and returns the
' un-escaped version of the string.
' Note that the only difference with the spec is that we can't escape \" in VBA
' so we ignore this particular case.
' In .Net, escaped sequences in strings are converted at compile time, so we end-up
' with an issue when trying to use our StringBuilder class or sFormat when they contain
' special escape chars such as \n or litteral unicode like \u20AC since they will not be
' converted to their string representation.
' This function solves this problem.
' See:
' - http://www.codeproject.com/Articles/371232/Escaping-in-Csharp-characters-strings-string-forma
' - http://msdn.microsoft.com/en-us/library/aa691087(v=vs.71).aspx
' - http://msdn.microsoft.com/en-us/library/aa664669(v=vs.71).aspx
'-----------------------------------------------------------------------------
Public Function UnEscStr(ByRef str As Variant) As String
    If VarType(str) <> vbString Then Exit Function

    ' To make conversion easy and fast, we store converted character litterals
    ' in a static array so we only need to construct this once.
    Static escVals(10) As String
    Static isInitialised As Boolean
    If Not isInitialised Then
        isInitialised = True
        escVals(1) = "'"
        escVals(2) = "\"
        escVals(3) = ChrW$(0)    ' Null
        escVals(4) = ChrW$(7)    ' Alert
        escVals(5) = ChrW$(8)    ' Backspace
        escVals(6) = ChrW$(&HC)  ' Form feed
        escVals(7) = vbLf        ' New line
        escVals(8) = vbCr        ' Carriage return
        escVals(9) = vbTab       ' Tab
        escVals(10) = ChrW$(&HB) ' Vertical tab
    End If
    
    Dim start As Long, pos As Long, i As Long, u As Long
    Dim c As String, v As String
    Dim l As Integer
    start = 1
    ' We'll search for each '\' and
    Do
        pos = InStr(start, str, "\", vbBinaryCompare)
        
        ' No '\' found, we're finished with this string
        If pos = 0 Then
            If start = 1 Then
                ' There was nothing to un-escape at all in the original
                ' string, so we just return it verbatim.
                UnEscStr = str
            Else
                ' Append the final bit of the original string to the result
                UnEscStr = UnEscStr & Mid$(str, start)
            End If
            Exit Do
        End If
        
        ' Append the bit of string we found before the escaped sequence
        UnEscStr = UnEscStr & Mid$(str, start, pos - start)
        pos = pos + 1
        start = pos + 1
        c = Mid$(str, pos, 1)
        ' Search the escaped character in our list of valid ones
        i = InStr(1, "'\0abfnrtv", c, vbBinaryCompare)
        If i > 0 Then
            ' We found a standard character litteral, just append
            ' its converted version to the result
            UnEscStr = UnEscStr & escVals(i)
        Else
            'Deal with Unicode
            Select Case AscW(c)
                Case 85  ' U: UTF-32 defined unicode point
                    ' Must have exactly 8 digits
                    ' Get the whole character code point
                    u = CLng("&H" & Mid$(str, start, 8))
                    If u < &HFFFF& Then
                        ' Standard unicode point that fits into UTF-16
                        v = ChrW$(u)
                        UnEscStr = UnEscStr & v
                    ElseIf u < &H10FFFF Then
                        ' We need to convert the UTF-32 into to a UTF-16 surrogate pair
                        ' See calculation http://www.unicode.org/faq/utf_bom.html#utf16-4
                        v = ChrW$(&HD7C0& + (u \ 1024))
                        UnEscStr = UnEscStr & v
                        v = ChrW$(&HDC00& + (u And &H3FF&))
                        UnEscStr = UnEscStr & v
                    End If
                    start = start + 8
                    
                Case 117 ' u: Standard unicode point that fits into UTF-16
                    ' Must have exactly 4 digits
                    v = ChrW$(CLng("&H" & Mid$(str, start, 4)))
                    UnEscStr = UnEscStr & v
                    start = start + 4
                
                Case 120 ' x: Standard unicode point that fits into UTF-16
                    ' Variable length, 1 to 4 digits
                    pos = start: l = 0
                    Do While InStr(1, "0123456789ABCDEFabcdef", Mid$(str, pos, 1), vbBinaryCompare) > 0 And (l < 4)
                        l = l + 1
                        pos = pos + 1
                    Loop
                    v = ChrW(CLng("&H" & Mid$(str, start, l)))
                    UnEscStr = UnEscStr & v
                    start = start + l
                
                Case Else
                    ' Unknown escape sequence, just leave it as is in the result
                    UnEscStr = UnEscStr & "\" & c
            End Select
        End If
    Loop While True
End Function



'-----------------------------------------------------------------------------
' Convert the given Variant into a base Type.
' For instance, if we pass a DAO.Field with subtype Integer,
' ConvertToBase() will return a variant of subtype Integer.
' Emty, Null are converted to emtpy strings or
'-----------------------------------------------------------------------------
Public Function ConvertVariantToBaseType(ByRef value As Variant) As Variant
    If Not IsObject(value) Then
        ConvertVariantToBaseType = value
        Exit Function
    End If
    'Detect if we're dealing with an array
    If VarType(value) And &H2000 Then
        ConvertVariantToBaseType = "<Array of " & TypeName(value) & " >"
        Exit Function
    End If
    
    Select Case VarType(value)
        Case VbVarType.vbString
            ConvertVariantToBaseType = CStr(value)
        Case VbVarType.vbBoolean
            ConvertVariantToBaseType = CBool(value)
        Case VbVarType.vbByte
            ConvertVariantToBaseType = CByte(value)
        Case VbVarType.vbCurrency
            ConvertVariantToBaseType = CCur(value)
        Case VbVarType.vbDate
            ConvertVariantToBaseType = CDate(value)
        Case VbVarType.vbDecimal
            ConvertVariantToBaseType = CDec(value)
        Case VbVarType.vbDouble
            ConvertVariantToBaseType = CDbl(value)
        Case VbVarType.vbInteger
            ConvertVariantToBaseType = CInt(value)
        Case VbVarType.vbLong
            ConvertVariantToBaseType = CLng(value)
        Case VbVarType.vbNull
            ConvertVariantToBaseType = vbNullString
        Case VbVarType.vbEmpty
            ConvertVariantToBaseType = vbNullString
        Case VbVarType.vbObject
            ConvertVariantToBaseType = "<object:" & TypeName(value) & ">"
        Case VbVarType.vbError
            ConvertVariantToBaseType = "<Error:" & value & ">"
        Case VbVarType.vbSingle
            ConvertVariantToBaseType = CSng(value)
        Case VbVarType.vbUserDefinedType
            ConvertVariantToBaseType = "<user-defined-type>"
        #If Win64 Then
        Case VbVarType.LongLong
            ConvertVariantToBaseType = CLngLng(value)
        #End If
        Case Else
            ConvertVariantToBaseType = "<Type:" & TypeName(value) & ">"
    End Select
End Function
